module idrs
  ! IDRS = Induced Dimension Reduction Method
  !
  ! Theory of this Krylov solver developed by :
  !
  ! PETER SONNEVELD AND MARTIN B. VAN GIJZEN
  !
  ! Fortran90 version adopted from:
  ! MATLAB file written by
  ! Martin van Gijzen and Peter Sonneveld
  ! Version August 31, 2010
  !
  ! For original MATLAB source code refer to:
  !
  ! http://ta.twi.tudelft.nl/nw/users/gijzen/IDR.html  
  !
  ! The software is distributed without any warranty.
  !
  ! Fortran version by Arash Ghasemi,
  !     ghasemi.arash@gmail.com
  ! Copyright (c) December 2014
  ! Version 1.0  
  !
  use normal_rand
  implicit none

  private

  public :: idrs_advanced, lin_solve

contains

  !
  !               The Main IDRS Solver
  !
  ! This is the improved version of original SIAM paper
  ! which has smoothing and preconditiong remedies.
  !
  subroutine idrs_advanced(A, b, s, tol, maxit, M1, M2, x0 &
       , smoothing, x, flag, resvec, iter, relres)
    implicit none
    real*8, dimension(:, :), intent(in) :: A
    real*8, dimension(:), intent(in) :: b
    integer, intent(in) :: s
    real*8, intent(in) :: tol
    integer, intent(in) :: maxit
    real*8, dimension(:, :), intent(in) :: M1, M2
    real*8, dimension(:), intent(in) :: x0
    integer, intent(in) :: smoothing

    ! outputs
    real*8, dimension(:), intent(out) :: x
    integer, intent(out) :: flag
    real*8, dimension(:), allocatable :: resvec
    integer, intent(out) :: iter
    real*8, intent(out) :: relres

    ! local vars
    integer :: k, i
    integer :: m, n
    real*8 :: angle, mp, normb, tolb, normr, trueres, om, alpha, beta, gamma
    real*8, dimension(size(A, 2), s) :: P, Q, G, U
    real*8, dimension(s, s) :: MM
    real*8, dimension(size(b), 1) :: r, x_s, r_s, v, tmp, t
    real*8, dimension(s, 1) :: f, c 
    integer ( kind = 4 ) :: seed

    ! HARD init
    x = 0.0d0
    flag = 0
    iter = 0
    relres = 0.0d0

    m = size(A, 1)
    n = size(A, 2)

    if (m .ne. n) then
       print *, 'Matrix must be square in idrs_advanced(...)! stop!'
       stop
    end if

    angle = 0.7d0
    seed = 123456789
    call r8mat_normal_01 ( n, s, seed, P )
    call orth(P, Q)
    P = Q

    ! % Number close to machine precision:
    mp = 1.0d3 * epsilon(1.0d0)

    relres = 0.0d0
    relres = relres / relres ! gets NaN by a trick

    ! Compute initial residual:
    x = x0
    normb = norm(b)
    tolb = tol * normb   ! Relative tolerance
    r(:,1) = b - matmul(A, x)

    if (smoothing .ne. 0) then
       x_s(:, 1) = x0
       r_s = r
    end if

    normr = norm(r(:,1))
    if ( allocated(resvec) ) deallocate(resvec) ! just for the 1st time
    call pushit(resvec, normr)
    trueres = 0.0d0

    if (normr <= tolb) then  !Initial guess is a good enough solution
       iter = 0                 
       flag = 0
       relres = normr/normb
       return
    end if
    ! %
    G = 0.0d0; U = 0.0d0; MM = 0.0d0
    do i = 1, s
       MM(i, i) = 1.0d0
    end do
    om = 1.0d0
    !

    ! % Main iteration loop, build G-spaces:
    iter = 0
    do ! while ( normr > tolb && iter < maxit )  

       ! % New righ-hand size for small system:
       f = transpose(matmul(transpose(r), P))
       do k = 1, s 
          !
          ! % Solve small system and make v orthogonal to P:
          ! % Note that MM is lower triangular, so a forward substitution would have been enough
          call lin_solve(MM(k:s,k:s), f(k:s,1), c(k:s,:))
          v = r - matmul(G(:,k:s), c(k:s,:)) 
          ! %
          ! % Preconditioning:
          call lin_solve(M1, v(:,1), tmp)
          v = tmp
          call lin_solve(M2, v(:,1), tmp)
          v = tmp

          ! % Compute new U(:,k) and G(:,k), G(:,k) is in space G_j
          U(:,k) = matmul(U(:,k:s), c(k:s,1)) + om * v(:, 1)
          G(:,k) = matmul(A, U(:,k))

          ! %
          ! % Bi-Orthogonalise the new basis vectors: 
          do i = 1, (k-1)
             alpha =  sum(P(:,i) * G(:,k)) / MM(i,i)
             G(:,k) = G(:,k) - alpha*G(:,i)
             U(:,k) = U(:,k) - alpha*U(:,i)
          end do
          ! %
          ! % New column of MM = P'*G  (first k-1 entries are zero)
          MM(k:s,k:k) = transpose(matmul(transpose(G(:,k:k)), P(:,k:s)))
          if ( MM(k,k) .eq. 0.0d0 ) then
             flag = 3
             return
          end if
          ! %
          ! %  Make r orthogonal to q_i, i = 1..k 
          beta = f(k,1)/MM(k,k)
          r = r - beta*G(:,k:k)
          x = x + beta*U(:,k)
          normr = norm(r(:, 1))

          ! %
          ! %  Smoothing:
          if ( smoothing .ne. 0) then
             t = r_s - r
             gamma =  sum(t*r_s)/sum(t*t)
             r_s = r_s - gamma*t
             x_s(:, 1) = x_s(:, 1) - gamma*(x_s(:, 1) - x)
             normr = norm(r_s(:,1))
          end if
          call pushit(resvec, normr)
          iter = iter + 1
          if ( (normr < tolb) .or. (iter .eq. maxit ) ) exit
          ! %
          ! % New f = P'*r (first k  components are zero)
          if ( k < s ) then 
             f((k+1):s, 1)   = f((k+1):s, 1) - beta*MM((k+1):s,k)
          end if
       end do
       ! %
       if ( (normr < tolb) .or. (iter .eq. maxit) ) then
          exit
       end if
       ! %
       ! % Now we have sufficient vectors in G_j to compute residual in G_j+1
       ! % Note: r is already perpendicular to P so v = r
       ! %
       ! % Preconditioning:
       v = r
       call lin_solve(M1, v(:,1), tmp)
       v = tmp
       call lin_solve(M2, v(:,1), tmp)
       v = tmp

       ! %
       ! % Matrix-vector multiplication:
       t = matmul(A, v)

       ! % Computation of a new omega
       om = omega( t(:, 1), r(:, 1), angle )
       if ( om .eq. 0.0d0 ) then
          flag = 3
          return
       end if
       ! %
       r = r - om*t
       x = x + om*v(:,1)
       normr = norm(r(:,1))
       ! %
       ! %     Smoothing:
       if ( smoothing .ne. 0 ) then
          t    = r_s - r
          gamma = sum(t*r_s)/sum(t*t)
          r_s = r_s - gamma*t
          x_s(:,1) = x_s(:,1) - gamma*(x_s(:,1) - x)
          normr = norm(r_s(:,1))
       end if
       ! %
       call pushit(resvec, normr)
       iter = iter + 1

       if ( (normr > tolb) .and. (iter < maxit) ) then
          cycle ! continue while loop
       else
          exit  ! exit while loop
       end if
    end do ! while

    if ( smoothing .ne. 0) then
       x = x_s(:,1)
    end if

    relres = norm(b - matmul(A, x)) / normb

    if ( relres < tol ) then 
       flag = 0
    else if ( iter .eq. maxit ) then
       flag = 1
    else
       flag = 2
    end if

    ! done here
  end subroutine idrs_advanced

  function omega( t, s, angle ) result(om)
    implicit none
    real*8, dimension(:), intent(in) :: t, s
    real*8, intent(in) :: angle
    real*8 :: om

    ! locals
    real*8 :: ns, nt, ts, rho

    ns = norm(s)
    nt = norm(t)
    ts = sum(t*s)
    rho = abs(ts/(nt*ns))
    om = ts/(nt*nt)
    if ( rho < angle ) then
       om = om*angle/rho
    end if

    ! done here
  end function omega

  ! generates orthonormal basis for range of
  ! a not neccessarily square matrix <A>
  subroutine orth(A, Q)
    implicit none
    real*8, dimension(:, :), intent(in)  :: A
    real*8, dimension(:, :), intent(out) :: Q

    ! local vars
    character*1 :: jobu, jobvt
    real*8, dimension(:, :), allocatable :: u, vt

    integer :: m, n, lda, lwork, INFO
    real*8, dimension(:), allocatable :: s, work

    ! write to temp; do not change original <A>
    Q = A

    ! init
    m = size(A, 1)
    n = size(A, 2)
    jobu = 'O'
    jobvt = 'N'
    lda = m
    allocate(s(min(m,n)))
    allocate(u(1, 1), vt(1, 1)) ! really not referenced!!!
    lwork = max(1,3*min(m,n)+max(m,n),5*min(m,n))
    allocate(work(max(1,lwork)))

    ! call LAPACK double precision generic SVD routine
    ! NOTE : svd is Overwritten in [Q]
    call dgesvd(jobu, jobvt, m, n, Q, lda, s, u, 1, vt, 1, work, lwork, INFO)

    if (INFO .ne. 0) then
       print *, 'something is wrong in SVD in orth(...)! stop'
       stop
    end if

    ! clean ups
    deallocate(u, vt)
    deallocate(s, work)

    ! done here
  end subroutine orth

  ! computes the L2 norm of 1d array <a>
  function norm(a)
    implicit none
    real*8, dimension(:), intent(in) :: a
    real*8 :: norm

    norm = sqrt(sum(a * a))

    ! done here
  end function norm

  ! adds the double precision val to the
  ! end of the array <a> 
  subroutine pushit(a, val)
    implicit none
    real*8, dimension(:), allocatable :: a
    real*8, intent(in) :: val

    ! local vals
    integer :: tsize, n
    real*8, dimension(:), allocatable :: tmp

    if ( .not. allocated(a) ) then 
       tsize = 0
    else 
       tsize = size(a)
    end if

    n = tsize + 1
    allocate(tmp(n))
    if (tsize >= 1) tmp(1:tsize) = a(1:tsize)
    tmp(n) = val

    call move_alloc(tmp, a)

    ! done here
  end subroutine pushit

  ! performs LAPACK general LU solve DGESV
  subroutine lin_solve(A, b, x)
    implicit none
    real*8, dimension(:, :), intent(in) :: A
    real*8, dimension(:), intent(in) :: b
    real*8, dimension(:,:), intent(out) :: x

    ! local vars
    integer :: N, NRHS, LDA, LDB, INFO
    real*8, dimension(size(A,1), size(A,1)) :: Atmp
    integer, dimension(size(A,1)) :: IPIV

    ! init
    N = size(A,1)
    NRHS = 1
    Atmp = A ! take copy; DONT modify the original one!
    LDA = N
    x(:,1) = b
    LDB = N
    INFO = 0
    call DGESV( N, NRHS, Atmp, LDA, IPIV, x, LDB, INFO )

    if ( INFO .ne. 0) then
       print *, 'something is wrong in lin_solve! stop'
       stop
    end if

    ! done here
  end subroutine lin_solve

end module idrs

! little program for testing
! this module
! comment after this line when
! using this module in your project
! by : Arash Ghasemi (ghasemi.arash@gmail.com)
program tester
  use normal_rand
  use idrs
  implicit none

  ! local vars
  integer :: n, i
  integer (kind = 4) :: seed
  integer :: s, maxit, smoothing, flag, iter
  real*8 :: tol, relres
  real*8, dimension(:, :), allocatable :: A, M1, M2
  real*8, dimension(:), allocatable :: b, x, x0, resvec
  real*8, dimension(:,:), allocatable :: xx

  ! init
  n = 100
  allocate(A(n,n), b(n), M1(n,n), M2(n,n), x(n), x0(n))
  allocate(xx(n, 1))

  seed = 123456789
  call r8mat_normal_01 ( n, n, seed, A )
  b = 1.0d0
  x0 = b
  M1 = 0.0d0
  M2 = 0.0d0
  do i = 1, n
     M1(i, i) = 1.0d0
     M2(i, i) = 1.0d0
  end do

  s = 10; maxit = 410; smoothing = 1; flag = 0; 
  tol = 1.0d-14

  call idrs_advanced(A, b, s, tol, maxit, M1, M2, x0 &
       , smoothing, x, flag, resvec, iter, relres)

  print *, 'flag = ', flag
  print *, 'resvec = ', resvec
  print *, 'iter = ', iter
  print *, 'relres = ', relres
  call lin_solve(A, b, xx)
  print *, 'Norm_Inf(xLU - xIDRS) = ', maxval(abs(xx(:,1) - x))

  ! clean ups
  deallocate(A, b, M1, M2, x, x0)
  deallocate(xx, resvec)

  ! done here
end program tester
